#$Header: /raid/cvsroot/DBIx/Record/Record.pm,v 1.46 2000/06/26 18:33:50 jesse Exp $
package DBIx::Record;

use strict;
use vars qw($VERSION @ISA $AUTOLOAD);
require Date::Kronos;

$VERSION = '0.02';

# Preloaded methods go here.

$DBIx::Record::Debug = 1; 

# {{{ sub new 

#instantiate a new record object.

sub new  {
    my $proto = shift;
    my $class = ref($proto) || $proto;
    my $self  = {};
    bless ($self, $class);
    return $self;
    # what are we doing here? TODO. why do we return early
    $self->_Init(@_);
  }

# }}}

# {{{ sub Id and id
sub Id  {
    my $self = shift;
    return ($self->{'values'}->{'id'});
  }

sub id  {
    my $self = shift;
    return ($self->Id);
  }

# }}}

# {{{ Datehandling

# There is room for optimalizations in most of those subs:

# {{{ LastUpdatedObj

sub LastUpdatedObj {
    my $self=shift;
    my $obj=Date::Kronos->new();
    $obj->sql_timestamp($self->LastUpdated);
    return $obj;
}

# }}}

# {{{ CreatedObj

sub CreatedObj {
    my $self=shift;
    my $obj=Date::Kronos->new();
    $obj->sql_timestamp($self->Created);
    return $obj;
}

# }}}

# {{{ AgeAsString
sub AgeAsString {
    my $self=shift;
    my $now=Date::Kronos->new(cal_type=>'Unix');
    my $age=$now-$self->CreatedObj;
    return "$age ago";
}
# }}}

# {{{ LastUpdatedAsString

sub LastUpdatedAsString {
    my $self=shift;
    if ($self->LastUpdated) {
	return $self->LastUpdatedObj->Gregorian->sql_timestamp;
    } else {
	return "never";
    }
}

# }}}

# {{{ CreatedAsString
sub CreatedAsString {
    return CreatedObj(@_)->Gregorian->stringify;
}
# }}}

# {{{ LongSinceUpdateAsString
sub LongSinceUpdateAsString {
    my $self=shift;
    if ($self->LastUpdated) {
	my $now=Date::Kronos->new(cal_type=>'Unix');
	my $age=$now-$self->LastUpdatedObj;
	return "$age ago";
    } else {
	return "never";
    }
}
# }}}

# }}} Datehandling

# {{{ Routines dealing with getting and setting row data

# {{{ sub DESTROY
sub DESTROY {
    1;
}
# }}}

# {{{ sub AUTOLOAD 
sub AUTOLOAD  {
  my $self = shift;
  
  no strict 'refs';

  if ($AUTOLOAD =~ /.*::(\w+)/ && $self->_Accessible($1,'read')) {
    my $Attrib = $1;

    *{$AUTOLOAD} = sub { return ($_[0]->_Value($Attrib))};
    return($self->_Value($Attrib));
  }
    
  elsif ($AUTOLOAD =~ /.*::Set(\w+)/ && $self->_Accessible($1,'write')) {
    my $Attrib = $1;

    *{$AUTOLOAD} = sub {  return ($_[0]->_Set($Attrib, $_[1]))};
    my $Value = shift @_;
    return($self->_Set($Attrib, $Value));
    }

  #Previously, I checked for writability here. but I'm not sure that's the
  #right idea. it breaks the ability to do ValidateQueue for a ticket
  #on creation.

  elsif ($AUTOLOAD =~ /.*::Validate(\w+)/ ) {
    my $Attrib = $1;

    *{$AUTOLOAD} = sub {  return ($_[0]->_Validate($Attrib, $_[1]))};
    my $Value = shift @_;
    return($self->_Validate($Attrib, $Value));
    }


# TODO: if autoload = 0 or 1 _ then a combination of lowercase and _ chars, 
# turn them into studlycapped phrases


 
  else {
    my ($package, $filename, $line);
    ($package, $filename, $line) = caller;

    die "$AUTOLOAD Unimplemented in $package. ($filename line $line) \n";
  }
  
}
# }}}

# {{{ sub _Accessible 

sub _Accessible  {
  my $self = shift;
  my $Attrib = shift;
  my $Mode = shift;
  my %Cols = @_;

  #  return true if we can $mode $Attrib;
  #return 0 if it's not a valid attribute;
  if (!$Cols{"$Attrib"}) { return(0)}
  # print  "Attrib: $Attrib: $Cols{$Attrib} '$Mode'\n";
  
  $Cols{$Attrib} =~ /$Mode/;
  
}

# }}}

# {{{ sub _Value 
sub _Value  {
    my $self = shift;
    my $field = shift;
    
#    if (defined($self->{'values'}->{$field})) {
      return($self->{'values'}->{"$field"});
#    }
#    else {
#      return('');
#    }
  }

# }}}

# {{{ sub _Set 

sub _Set  {
    my $self = shift;
    my $field = shift;
    my $value = shift; 
    my $is_sql = shift;
    my ($error_condition);
    #print "$self -> _Set $field->$value\n";
    
    defined $field && !defined($value) && return (0,"No value sent to _Set!\n");
    
    if (defined $field) {
	if (   (defined $self->_Value($field)) 
	    && ($value eq $self->_Value($field))
	       ) {
	    return (0, "That is already the current value");
	} else {
	    #TODO $self->_Validate($field, $value);
	    $error_condition = 
		$self->{'DBIxHandle'}->UpdateTableValue
		    ($self->{'table'}, $field,$value,$self->id, $is_sql);
	    # TODO: Deal with error handling?
	    $self->{'values'}->{"$field"} = $value;
	}
    }
    $error_condition = 
	$self->{'DBIxHandle'}->UpdateTableValue
	    ($self->{'table'}, 'LastUpdated','now()',$self->id)
		if ($self->_Accessible('LastUpdated','auto'));
    return (1, "The new value has been set.");
}

# }}}
# {{{ sub _Validate 

#TODO: Implement validate.
sub _Validate  {
    my $self = shift;
    my $field = shift;
    my $value = shift;
    
    
    #Check type of input
    #If it's null, are nulls permitted?
    #If it's an int, check the # of bits
    #If it's a string, 
    #check length
    #check for nonprintables
    #If it's a blob, check for length
    #In an ideal world, if this is a link to another table, check the dependency.
    
    
    
  }	

# }}}	

# }}}
  
# {{{ routines dealing with loading records

# {{{ sub Load 

# load should do a bit of overloading
# if we call it with only one argument, we're trying to load by reference.
# if we call it with a passel of arguments, we're trying to load by value
# The latter is primarily important when we've got a whole set of record that we're
# reading in with a recordset class and want to instantiate objefcts for each record.

*load =\&Load;

sub Load  {
    my $self = shift;
    my ($package, $filename, $line) = caller;
    return $self->LoadById(@_);
  }
  
  # TODO
  # load by value is intended to be called when it's loading a 
  # bunch of attestations. it takes a hash of the values of the record and
  # instantiates the object from that. 

# }}}

# {{{ sub load_by_col 

*load_by_col=\&LoadByCol;

# }}}

# {{{ sub LoadByCol 

sub LoadByCol  {
    my $self = shift;
    my $col = shift;
    my $val = shift;
    
    $val = $self->_Handle->quote($val);
    my $QueryString = "SELECT  * FROM ".$self->{'table'}." WHERE $col = $val";
    return ($self->_LoadFromSQL($QueryString));
  }

# }}}

# {{{ sub LoadById 
sub LoadById  {
    my $self = shift;
    my $id = shift;

    $id = 0 if (!defined($id));
    return ($self->LoadByCol('id',$id));
}
# }}}  


# {{{ sub LoadFromHash
sub LoadFromHash {
  my $self = shift;
  my $hashref = shift;
  
  $self->{'values'} = $hashref;
  return ($self->{'values'}{'id'});
}
# }}}

# {{{ sub _LoadFromSQL 

sub _LoadFromSQL  {
    my $self = shift;
    my $QueryString = shift;
    
    my ($hash, $sth);
    $sth = $self->_Handle->prepare($QueryString);
    
    if (!$sth) {
      die "Error:" . $self->_Handle->errstr . "\n";
    }

    if (!$sth->execute) {
	warn "some problem here";
	return(undef);
    }

    #TODO: COMPATIBILITY PROBLEM with fetchrow_hashref!
    #Some DBMS'es returns uppercase, some returns lowercase,
    #and mysql return mixedcase!

    #TODO this only gets the first row. we should check if there are more.
    $self->{'values'} = $sth->fetchrow_hashref;
    unless ($self->{'values'}) {
	warn "something might be wrong here; row not found. SQL: $QueryString";
	return undef;
    }
    #$self->{'values'}->{'id'} = $id;
    unless ($self->{'values'}{'id'}) {
	warn "something wrong here";
    }
    return ($self->{'values'}{'id'});
  }

# }}}

# }}}

# {{{ Routines dealing with creating or deleting rows in the DB

# {{{ sub Create 

sub Create  {
    my $self = shift;
    my @keyvalpairs = (@_);

    push @keyvalpairs, 'Created', 'now()'
	if $self->_Accessible('Created', 'auto');
    
    my ($QueryString, $value, $key, $cols, $vals);
    
    while ($key = shift @keyvalpairs) {
      $value = shift @keyvalpairs;
    
      $cols .= $key . ", ";
      if (defined ($value)) {
	  $value = $self->{'DBIxHandle'}->safe_quote($value)
	      unless ($key=~/^(Created|LastUpdated)$/ && $value=~/^now\(\)$/i);
	  $vals .= "$value, ";
      }
      else {
	$vals .= "NULL, ";
      }
    }	
    
    $cols =~ s/, $//;
    $vals =~ s/, $//;
    #TODO Check to make sure the key's not already listed.
    #TODO update internal data structure
    $QueryString = "INSERT INTO ".$self->{'table'}." ($cols) VALUES ($vals)";

    #print STDERR "$QueryString\n";

    my $sth = $self->_Handle->prepare($QueryString);
    if (!$sth) {
      
      if ($main::debug) {
	die "Error:" . $self->_Handle->errstr . "\n";
      }
      else {
	return (0);
	
      }
    }
    if (!$sth->execute) {
      if ($main::debug) {
	die "Error:" . $sth->errstr . "\n";
      }
      else {
	return(0);
      }
    }
    
    $self->{'id'}=$sth->{'mysql_insertid'};
    return( $self->{'id'}); #Add Succeded. return the id
  }

# }}}

# {{{ sub delete 
sub delete  {
  my $self = shift;
  return ($self->Delete(@_));
}
# }}}

# {{{ sub Delete 

sub Delete  {
    my $self = shift;
    
    my $QueryString;
    
    #TODO Check to make sure the key's not already listed.
    #TODO Update internal data structure
    $QueryString = "DELETE FROM ".$self->{'table'} . " WHERE id  = ". $self->id();
    my $sth = $self->_Handle->prepare($QueryString);
    if (!$sth) {
      if ($main::debug) {
	die "Error:" . $self->_Handle->errstr . "\n";
      }
      else {
	return (0);
      }
    }
    if (!$sth->execute) {
      if ($main::debug) {
	die "Error:" . $sth->errstr . "\n";
      }
      else {
	return(0);
      }

    }
    
    return (1);		#Update Succeded
  }

# }}}

# }}}

# {{{ Routines dealing with database handles

# TODO: Combine into _Handle this will involve modifying subclasses.

# {{{ sub _MyHandle 
sub _MyHandle  {
    my $self = shift;
    my %args = ( Handle=> undef,
		 @_
	       );
    $self->{'DBIxHandle'} = $args{'Handle'};

  }
# }}}

# {{{ sub _Handle 
sub _Handle  {
  my $self = shift;
  return ($self->{'DBIxHandle'}->Handle);
  
}
# }}}

# }}}

1;

__END__

# {{{ POD

=head1 NAME

DBIx::Record - Perl extension for subclassing, so you can deal with a Record

=head1 SYNOPSIS

  use DBIx::Record;


=head1 DESCRIPTION
DBIX::Record is designed to work with DBIx::EasySearch. Users should almost
never create DBIx::Record objects themselves.

Docs are forthcoming. If you pester jesse@fsck.com he'll put them together.

Check out Request Tracker version 1.1 or newer at http://www.fsck.com/projects/rt/ for examples of usage.

=head1 AUTHOR

Jesse Vincent, jesse@fsck.com

=head1 SEE ALSO

perl(1).

=cut

# }}}

